/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "ors_control.h"
#include "ors_swift.h"

ControlVariableList globalSpace;

//===========================================================================
//
// CVs
//

ControlVariable::ControlVariable(){
  globalSpace.memMove=true;
  globalSpace.append(this);
  active=false;
  type=noneCVT;
  targetType=noneTT;
  prec=0.; vprec=0.; a=b=0.; state=-1; state_tol=.05; err=derr=0.;
}

ControlVariable::ControlVariable(
    const char* _name,
    ors::Graph& _sl,
    CVtype _type,
    const char *iname,const char *iframe,
    const char *jname,const char *jframe,
    const arr& _params){
  globalSpace.memMove=true;
  globalSpace.append(this);
  active=false;
  type=noneCVT;
  targetType=noneTT;
  prec=0.; vprec=0.; a=b=0.; state=-1; state_tol=.05;
  set(
    _name,_sl,_type,
    iname  ? (int)_sl.getName(iname)->index      : -1,
    iframe ? ors::Frame().setText(iframe) : ors::Frame(),
    jname  ? (int)_sl.getName(jname)->index      : -1,
    jframe ? ors::Frame().setText(jframe) : ors::Frame(),
    _params);
}

ControlVariable::~ControlVariable(){
  globalSpace.removeValue(this);
}

void ControlVariable::set(
    const char* _name,
    ors::Graph &_ors,
    CVtype _type,
    int _i,const ors::Frame& _irel,
    int _j,const ors::Frame& _jrel,
    const arr& _params){
  type=_type;
  name=_name;
  ors=&_ors;
  i=_i;
  irel=_irel;
  j=_j;
  jrel=_jrel;
  params=_params;
  updateState();
  updateJacobian();
  x_target=x;
  v_target=v;
}

/*void ControlVariable::set(const char* _name,ors::Graph& _sl,CVtype _type,const char *iname,const char *jname,const char *reltext){
  set(
    _name,_sl,_type,
    _sl.getName(iname)->index,
    _sl.getName(jname)->index,
    ors::Frame().setText(reltext));
}*/

void ControlVariable::setGains(double _a,double _b){
  targetType=gainsTT;
  active=true;
  a=_a;
  b=_b;
  if(!prec) prec=100.;
}

void ControlVariable::setGainsAsAttractor(double decaySteps,double oscillations){
  targetType=attractorTT;
  active=true;
  b=2./decaySteps;
  a=b*b*(MT_PI*MT_PI*oscillations*oscillations + .25);
  if(!prec) prec=100.;
}

//compute an x_trajectory and prec_trajectory which connects x with x_target and 0 with prec
void ControlVariable::setTrajectory(uint T,double funnelsdv,double funnelvsdv){
  targetType=trajectoryTT;
  active=true;
  uint t;
  double a;
  x_trajectory.resize(T,x.N);
  prec_trajectory.resize(T);
  v_trajectory.resize(T,x.N);
  vprec_trajectory.resize(T);
  for(t=0;t<T;t++){
    a = (double)t/(T-1);
    x_trajectory[t]()  = (1.-a)*x + a*x_target;
    prec_trajectory(t) = 1./MT::sqr(sqrt(1./prec) + (1.-a)*funnelsdv);

    v_trajectory[t]()  = (1.-a)*v + a*v_target;
    vprec_trajectory(t) = 1./MT::sqr(sqrt(1./vprec) + (1.-a)*funnelvsdv);
  }
}

//compute an x_trajectory and prec_trajectory which connects x with x_target and 0 with prec
void ControlVariable::setConstantTargetTrajectory(uint T){
  targetType=trajectoryTT;
  active=true;
  uint t;
  x_trajectory.resize(T,x.N);
  v_trajectory.resize(T,x.N);
  for(t=0;t<T;t++){
    x_trajectory[t]()  = x_target;
    v_trajectory[t]()  = v_target;
  }
}

//compute an x_trajectory and prec_trajectory which connects x with x_target and 0 with prec
void ControlVariable::setInterpolatedTargetTrajectory(uint T){
  targetType=trajectoryTT;
  active=true;
  uint t;
  double a;
  x_trajectory.resize(T,x.N);
  v_trajectory.resize(T,x.N);
  for(t=0;t<T;t++){
    a = (double)t/(T-1);
    x_trajectory[t]()  = (1.-a)*x + a*x_target;
    v_trajectory[t]()  = (1.-a)*v + a*v_target;
  }
}

void ControlVariable::setInterpolatedTargetsEndPrecisions(uint T,double inter_yprec,double end_yprec,double inter_vprec,double end_vprec){
  targetType=trajectoryTT;
  active=true;
  uint t;
  double a;
  x_trajectory.resize(T,x.N);    prec_trajectory.resize(T);
  v_trajectory.resize(T,x.N);   vprec_trajectory.resize(T);
  for(t=0;t<T;t++){
    a = (double)t/(T-1);
    x_trajectory[t]()  = (1.-a)*x + a*x_target;
    v_trajectory[t]()  = (1.-a)*v + a*v_target;
  }
  for(t=0;t<T-1;t++){
     prec_trajectory(t) = inter_yprec;
    vprec_trajectory(t) = inter_vprec;
  }
   prec_trajectory(T-1) = end_yprec;
  vprec_trajectory(T-1) = end_vprec;
}

void ControlVariable::setInterpolatedTargetsConstPrecisions(uint T,double yprec,double vprec){
  targetType=trajectoryTT;
  active=true;
  uint t;
  double a;
  x_trajectory.resize(T,x.N);    prec_trajectory.resize(T);
  v_trajectory.resize(T,x.N);   vprec_trajectory.resize(T);
  for(t=0;t<T;t++){
    a = (double)t/(T-1);
    x_trajectory[t]()  = (1.-a)*x + a*x_target;
    v_trajectory[t]()  = (1.-a)*v + a*v_target;
  }
  for(t=0;t<T;t++){
     prec_trajectory(t) = yprec;
    vprec_trajectory(t) = vprec;
  }
}

//compute an x_trajectory and prec_trajectory which connects x with x_target and 0 with prec
void ControlVariable::setPrecisionTrajectoryFinal(uint T,double intermediate_prec,double final_prec){
  active=true;
  uint t;
  prec_trajectory.resize(T);
  for(t=0;t<T-1;t++) prec_trajectory(t) = intermediate_prec;
  prec_trajectory(T-1) = final_prec;
}

//compute an x_trajectory and prec_trajectory which connects x with x_target and 0 with prec
void ControlVariable::setPrecisionTrajectoryConstant(uint T,double const_prec){
  active=true;
  prec_trajectory.resize(T);
  prec_trajectory = const_prec;
}

void ControlVariable::setPrecisionVTrajectoryFinal(uint T,double intermediate_vprec,double final_vprec){
  active=true;
  uint t;
  vprec_trajectory.resize(T);
  for(t=0;t<T-1;t++) vprec_trajectory(t) = intermediate_vprec;
  vprec_trajectory(T-1) = final_vprec;
}

//compute an x_trajectory and prec_trajectory which connects x with x_target and 0 with prec
void ControlVariable::setPrecisionVTrajectoryConstant(uint T,double const_prec){
  active=true;
  vprec_trajectory.resize(T);
  vprec_trajectory = const_prec;
}

void ControlVariable::updateState(){
  arr p;
  arr q,qv;
  ors::Vector pi,pj,c;
  arr zi,zj;
  ors::Frame f,fi,fj;

  v_old=v;
  x_old=x;
  
  //get state
  switch(type){
  case posCVT:
    if(j==-1){ ors->kinematics(x,i,&irel); break; }
    pi = ors->bodies(i)->X.p + ors->bodies(i)->X.r * irel.p;
    pj = ors->bodies(j)->X.p + ors->bodies(j)->X.r * jrel.p;
    c = ors->bodies(j)->X.r / (pi-pj);
    x.resize(3); x.setCarray(c.v,3);
    break;
  case zoriCVT:
    if(j==-1){ ors->kinematicsZ(x,i,&irel); break; }
    //relative
    MT_MSG("warning - don't have a correct Jacobian for this CVType yet");
    fi = ors->bodies(i)->X; fi.addRelativeFrame(irel);
    fj = ors->bodies(j)->X; fj.addRelativeFrame(jrel);
    f.setDifference(fi,fj);
    f.r.getZ(c);
    x.setCarray(c.v,3);
    break;
  case rotCVT:       x.resize(3); x.setZero(); break; //the _STATE_ of rot is always zero... the Jacobian not... (hack)
  case contactCVT:   ors->getPenetrationState(p); x.resize(1);  x(0) = p(i);  break;
  case gripCVT:      ors->getGripState(x,i);      break;
  case qLinearCVT:   ors->getJointState(q,qv);    x = params * q;   break;
  case qSquaredCVT:  ors->getJointState(q,qv);    x.resize(1);  x(0) = scalarProduct(params,q,q);  break;
  case qSingleCVT:   ors->getJointState(q,qv);    x.resize(1);  x(0)=q(-i);  break;
  case qLimitsCVT:   ors->getLimitsMeasure(x,params);    break;
  case comCVT:       ors->getCenterOfMass(x);     x.resizeCopy(2);  break;
  case collCVT:      ors->getContactMeasure(x,params(0));   break;
  case skinCVT:
    x.resize(params.N);
    x.setZero();
    break;
  case zalignCVT:
    ors->kinematicsZ(zi,i,&irel);
    if(j==-1){ zj.resize(3); zj.setZero(); zj(2)=1.; }
    else ors->kinematicsZ(zj,j,&jrel);
    x.resize(1);
    x(0) = scalarProduct(zi,zj);
    break;
  default:  HALT("no such CVT");
  }

  if(x_old.N!=x.N){
    x_old=x;
    v.resizeAs(x); v.setZero();
    v_old=v;
  }

  //v = .5*v + .5*(x - x_old);
  v = x - x_old;

  if(x_target.N==x.N){
    err=norm(x - x_target);
    derr=err - norm(x_old - x_target);
    if(err < state_tol && fabs(derr) < state_tol){
      state = 1;
    }else{
      state = 0;
    }
  }
}
  
void ControlVariable::updateJacobian(){
  arr q,qv;
  ors::Vector normal,d,vi,vj,r,jk,pi,pj;
  arr zi,zj,Ji,Jj,JRj;
  uint k;

  switch(type){
  case posCVT:
    if(j==-1){ ors->jacobian(J,i,&irel); break; }
    pi = ors->bodies(i)->X.p + ors->bodies(i)->X.r * irel.p;
    pj = ors->bodies(j)->X.p + ors->bodies(j)->X.r * jrel.p;
    ors->jacobian(Ji,i,&irel);
    ors->jacobian(Jj,j,&jrel);
    ors->jacobianR(JRj,j);
    J.resize(3,Jj.d1);
    for(k=0;k<Jj.d1;k++){
      vi.set(Ji (0,k),Ji (1,k),Ji (2,k));
      vj.set(Jj (0,k),Jj (1,k),Jj (2,k));
      r .set(JRj(0,k),JRj(1,k),JRj(2,k));
      jk =  ors->bodies(j)->X.r / (vi - vj);
      jk += ors->bodies(j)->X.r / ((-r) ^ (pi - pj));
      J(0,k)=jk(0); J(1,k)=jk(1); J(2,k)=jk(2); 
    }
    break;
  case zoriCVT:  ors->jacobianZ(J,i,&irel);   break;
  case rotCVT:   ors->jacobianR(J,i);   break;
  case contactCVT:  NIY;  break;
  case gripCVT:  NIY;  break;
  case qLinearCVT:  J = params;   break;
  case qSquaredCVT:
    ors->getJointState(q,qv);
    J = params * q;
    J *= 2.;
    J.reshape(1,q.N);
    break;
  case qSingleCVT:
    J.resize(1,ors->getJointStateDimension());
    J.setZero();
    J(0,-i) = 1.;
    break;
  case qLimitsCVT:   ors->getLimitsGradient(J,params);  break;
  case comCVT:    ors->getComGradient(J);  J.resizeCopy(2,J.d1);  break;
  case collCVT:   ors->getContactGradient(J,params(0));  break;
  case skinCVT:
    J.clear();
    uint k,l;
    for(k=0;k<params.N;k++){
      l=(uint)params(k);
      ors->jacobian(Ji,l,NULL);
      ors->bodies(l)->X.r.getY(vi);
      vi *= -1.;
      zi.setCarray(vi.v,3);
      J.append(zi*Ji);
    }
    J.reshape(params.N,J.N/params.N);
    break;
  case zalignCVT:
    ors->kinematicsZ(zi,i,&irel);
    ors->jacobianZ(Ji,i,&irel);
    if(j==-1){
      zj.resize(3);
      zj.setZero();
      zj(2)=1.;
      Jj.resizeAs(Ji);
      Jj.setZero();
    }else{
      ors->kinematicsZ(zj,j,&jrel);
      ors->jacobianZ(Jj,j,&jrel);
    }
    J = ~zj * Ji + ~zi * Jj;
    J.reshape(1,ors->jd);
    break;
  default:  NIY;
  }
  transpose(tJ,J);
}

void ControlVariable::getHessian(arr& H){
  switch(type){
  case posCVT:
    if(j==-1){ ors->hessian(H,i,&irel); break; }
  default:  NIY;
  }
}

void ControlVariable::updateChange(int t){
  CHECK(x.N,"variable needs to be updated before!");
  arr xt,vt;
  if(t!=-1){
    xt.referToSubDim(x_trajectory,t);
    vt.referToSubDim(v_trajectory,t);
    //prec     = prec_trajectory(t);
  }else{
    xt.referTo(x_target);
    vt.referTo(v_target);
  }
  CHECK(xt.N==x.N,"targets need to be set before");
  CHECK(vt.N==v.N,"targets need to be set before");
  switch(targetType){
  case directTT:{
    x_change = xt - x;
    break;
  }
  case gainsTT:{
    x_change = a*(xt - x) + b*(vt - v);
    break;
  }
  case attractorTT:{
    x_change = a*(xt - x) + b*(vt - v);
    x_change = v + x_change;
    break;
  }
  case trajectoryTT:{
    x_change = xt - x;
    break;
  }
  default:
    HALT("needs a target type! set targets before!");
  }
}

/*  switch(type){
  case zoriCVT:
    dx.resize(3);
    normal.set(x.p);
    d.set(x_change.p);
    d.makeNormal(normal);
    dx.setCarray(d.v,3);
    break;
  default:
    dx = x_change;
  default:
    break;
  }
    */

void ControlVariable::write(ostream &os) const{
  os <<"CV `" <<name;
  switch(type){
  case posCVT:     os <<"  (pos " <<ors->bodies(i)->name <<")"; break;
  //case relPosCVT:  os <<"  (relPos " <<ors->bodies(i)->name <<'-' <<ors->bodies(j)->name <<")"; break;
  case zoriCVT:    os <<"  (zori " <<ors->bodies(i)->name <<")"; break;
  case rotCVT:     os <<"  (rot " <<ors->bodies(i)->name <<")"; break;
  case contactCVT: os <<"  (contact " <<ors->bodies(i)->name <<' '<<params(0) <<")"; break;
  case gripCVT:    os <<"  (grip " <<ors->bodies(i)->name <<")"; break;
  case qLinearCVT: os <<"  (qLinear " <<sum(params) <<")"; break;
  case qSquaredCVT:os <<"  (qSquared " <<sum(params) <<")"; break;
  case qSingleCVT: os <<"  (qSingle " <<ors->joints(-i)->from->name <<'-' <<ors->joints(-i)->to->name <<")"; break;
  case qLimitsCVT: os <<"  (qLimitsCVT " <<sum(params) <<")"; break;
  case comCVT:     os <<"  (COM)"; break;
  case collCVT:    os <<"  (COLL)"; break;
  case zalignCVT:  os <<"  (zalign " <<ors->bodies(i)->name <<'-' <<ors->bodies(j)->name <<")"; break;
  default: HALT("CV::write - no such CVT");
  }
  os
    <<"\n  x=" <<x
    <<"\n  v=" <<v
    <<"\n  x_target=" <<x_target
    <<"\n  v_target=" <<v_target
    <<"\n  change=" <<x_change
    <<"\n  prec=" <<prec
    <<"\n  a=" <<a <<"  b=" <<b
    <<"\n  state=" <<state
    <<endl;
}

//===========================================================================
//
// ControlVariableList functions
//






void reportAll(ControlVariableList& CS,ostream& os,bool onlyActives){
  for(uint i=0;i<CS.N;i++) if(!onlyActives || CS(i)->active){
    os <<'[' <<i <<"] " <<*CS(i);
  }
}

void reportNames (ControlVariableList& CS,ostream& os,bool onlyActives){
  uint i,j,n=1;
  os <<"CVnames = {";
  for(i=0;i<CS.N;i++) if(!onlyActives || CS(i)->active){
    for(j=0;j<CS(i)->x.N;j++){
      os <<"'" <<n <<'-' <<CS(i)->name <<j <<"' ";
      n++;
    }
  }
  os <<"};" <<endl;
}

void reportState (ControlVariableList& CS,ostream& os,bool onlyActives){;
  uint i;
  MT::IOraw=true;
  for(i=0;i<CS.N;i++) if(!onlyActives || CS(i)->active){
    os <<CS(i)->x;
  }
  os <<endl;
}

void reportErrors(ControlVariableList& CS,ostream& os,bool onlyActives,int t){
  uint i;
  double e,E=0.;
  for(i=0;i<CS.N;i++) if(!onlyActives || CS(i)->active){
    if(t!=-1)
      if(t) e=norm(CS(i)->x - CS(i)->x_trajectory[t-1]);
      else  e=0.;
    else
      e=norm(CS(i)->x - CS(i)->x_target);
    os <<e <<' ';
    E += e; //*CS(i)->prec;
  }
  os <<E <<endl;
}

void updateState(ControlVariableList& CS){
  for(uint i=0;i<CS.N;i++){
    CS(i)->updateState();
  }
}

void updateJacobian(ControlVariableList& CS){
  for(uint i=0;i<CS.N;i++){
    CS(i)->updateJacobian();
  }
}

void updateChanges(ControlVariableList& CS,int t){
  for(uint i=0;i<CS.N;i++) if(CS(i)->active){
    CS(i)->updateChange(t);
  }
}

void getJointJacobian(ControlVariableList& CS,arr& J){
  uint i,n=0;
  J.clear();
  for(i=0;i<CS.N;i++) if(CS(i)->active){
    CS(i)->updateJacobian();
    J.append(CS(i)->J);
    n=CS(i)->J.d1;
  }
  J.reshape(J.N/n,n);
}

void getJointXchange(ControlVariableList& CS,arr& x_change){
  uint i;
  x_change.clear();
  for(i=0;i<CS.N;i++) if(CS(i)->active){
    x_change.append(CS(i)->x_change);
  }
  x_change.reshape(x_change.N);
}

double getCost_obsolete(ControlVariableList& CS,const arr& W,int t){
  uint i;
  double e,C=0.;
  for(i=0;i<CS.N;i++) if(CS(i)->active){
    if(t!=-1){
      e=sumOfSqr(CS(i)->x - CS(i)->x_trajectory[t]);
    }else{
      e=sumOfSqr(CS(i)->x - CS(i)->x_target);
    }
    C += e*CS(i)->prec;
    //cout <<"cost(" <<CS(i)->name <<") = " <<e <<"," <<e*CS(i)->prec <<endl;
  }
  return C;
}

void getCostGradient_obsolete(ControlVariableList& CS,arr& dCdq,const arr& W,int t){
  uint i,n=W.d0;
  dCdq.resize(n);
  dCdq.setZero();
  arr e,J,tJ,dx;
  for(i=0;i<CS.N;i++) if(CS(i)->active){
    CS(i)->updateJacobian();
    if(t!=-1){
      e=CS(i)->x - CS(i)->x_trajectory[t];
    }else{
      e=CS(i)->x - CS(i)->x_target;
    }
    dCdq += CS(i)->tJ * (e*(2.*CS(i)->prec));
  }
}

void hierarchicalControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W){
  uint i,n=W.d0;
  dq.resize(n);
  dq.setZero();
  arr hatJ,invHatJ,N;
  N.setId(n);
  arr invW;
  inverse_SymPosDef(invW,W);
  for(i=0;i<CS.N;i++) if(CS(i)->active){
    CS(i)->updateJacobian();

    hatJ = CS(i)->J * N;
    pseudoInverse(invHatJ,hatJ,invW,1e-5);
    dq += invHatJ * (CS(i)->x_change - CS(i)->J * dq);
    N  -= invHatJ * hatJ;
  }
}

void bayesianControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W){
  uint n=W.d0;
  dq.resize(n);
  dq.setZero();
  uint i;
  arr a(n),A(n,n),invA(n,n);
  //arr Q,JQ;
  A=W;
  a.setZero();
  arr w(3);
  for(i=0;i<CS.N;i++) if(CS(i)->active){
    CS(i)->updateJacobian();
    a += CS(i)->prec * CS(i)->tJ * CS(i)->x_change;
    A += CS(i)->prec * CS(i)->tJ * CS(i)->J;
  }
  inverse_SymPosDef(invA,A);
  dq = invA * a;
}

/*void bayesianIterateControl_obsolete(ControlVariableList& CS,
                            arr& qt,const arr& qt_1,const arr& W,double eps,uint maxIter){
  uint j;
  qt=qt_1;
  arr dq;
  for(j=0;j<maxIter;j++){
    top.setq(qt);
    bayesianIKControl(top,dq,W);
    if(j<3) qt+=dq;
    //else if(j<10) qt+=.8*dq;
    else qt+=.8*dq;
    if(dq.absMax()<eps) break;
  }
  if(j==maxIter) HALT("warning: IK didn't converge (|last step|="<<dq.absMax()<<")");
  else cout <<"IK converged after steps=" <<j <<endl;
}*/

void additiveControl_obsolete(ControlVariableList& CS,arr& dq,const arr& W){
  dq.resize(W.d0);
  dq.setZero();
  uint i,n=0;
  arr invJ;
  arr invW;
  inverse_SymPosDef(invW,W);
  for(i=CS.N;i--;) if(CS(i)->active){
    CS(i)->updateJacobian();
    pseudoInverse(invJ,CS(i)->J,invW,0.);
    dq += invJ * CS(i)->x_change;
    n++;
  }
  dq/=(double)n;
}

/*OLD
void bayesianPlanner_obsolete(ors::Graph *ors,ControlVariableList& CS,SwiftModule *swift,OpenGL *gl,
                     arr& q,uint T,const arr& W,uint iterations,
                     std::ostream* os,int display,bool repeat){
  //FOR THE OLD VERSION, SEE SMAC.CPP IN THE DEPOSIT
  uint n=W.d0,i;
  arr J,tJ,phiHatQ,dx;
  arr q0,qv0,invW;
  inverse_SymPosDef(invW,W);

  byteA img(300,500,3);

  ors->getJointState(q0,qv0);

  arr tmp,
    a(T,n),invA(T,n,n),
    z(T,n),invZ(T,n,n),
    b(T,n),B(T,n,n),invB(T,n,n),
    r(T,n),R(T,n,n),
    hatq(T,n);
  a[0]=q0;
  invA[0].setDiag(1e10);
  b[0]=q0;
  B[0].setDiag(1e-10);
  invB[0].setDiag(1e10);
  r[0]=0.;
  R[0]=0.;
  z.setZero();
  invZ.setZero();

  MT::timerStart();
  
  uint k,t,dt,t0;
  for(k=0;k<iterations;k++){
    if(!(k&1)){ dt=1; t0=1; }else{ dt=(uint)-1; t0=T-1; }
    for(t=t0;t<T && t>0;t+=dt){
      //compute (a,A)
      inverse_SymPosDef(tmp,invA[t-1] + R[t-1]);
      a[t] = tmp * (invA[t-1]*a[t-1] + r[t-1]);
      inverse_SymPosDef(invA[t](), invW + tmp);
      
      //cout <<"a\n" <<a[t] <<endl <<invA[t] <<endl;

      //compute (z,Z)
      if(k && t<T-1){
        inverse_SymPosDef(tmp,invZ[t+1] + R[t+1]);
        z[t] = tmp * (invZ[t+1]*z[t+1] + r[t+1]);
        inverse_SymPosDef(invZ[t](), invW + tmp);
      }
      if(k && t==T-1){
        z[t] = b[t];
        invZ[t].setDiag(1e10); //fixes the end posture!, use 1e-5 otherwise
      }

      //cout <<"z\n" <<z[t] <<endl <<invZ[t] <<endl;
      
      //compute (r,R)
      //if(k) hatq[t]()=.2*b[t]+.8*hatq[t]; else hatq[t]()=a[t];
      if(k) hatq[t]()=b[t]; else hatq[t]()=a[t];
      ors->setJointState(hatq[t]);
      ors->calcNodeFramesFromEdges();
      computeProxiesUsingSwift(*ors,*swift,false);
      //slGetProxies(*ors,*ode);
      r[t].setZero();
      R[t].setZero();
      for(i=0;i<CS.N;i++) if(CS(i)->active){
        CS(i)->updateState();    phiHatQ=CS(i)->x;
        CS(i)->updateJacobian(); J=CS(i)->J; tJ=CS(i)->tJ;
        dx = CS(i)->x_trajectory[t] - phiHatQ;
        r[t]() += CS(i)->prec_trajectory(t) * tJ * dx;
        R[t]() += CS(i)->prec_trajectory(t) * tJ * J;
      }
      r[t]() += R[t] * hatq[t];

      //cout <<"r\n" <<r[t] <<endl <<R[t] <<endl;

      //compute (b,B);
      invB[t] = invA[t] + invZ[t] + R[t];
      //cout <<"invB\n" <<invB[t] <<endl;
      inverse_SymPosDef(B[t](), invB[t]);
      b[t] = B[t] * (invA[t]*a[t] + invZ[t]*z[t] + r[t]);

      //cout <<"b\n" <<b[t] <<endl <<B[t] <<endl;

      //display
      if(display>0){
        ors->setJointState(b[t]);
        ors->calcNodeFramesFromEdges();
        //if(t==1 || !(t%display)){ gl->text.clr() <<k <<':' <<t; gl->update(); }
        //glGrabImage(img); write_ppm(img,STRING("imgs/plan_"<<std::setfill('0') <<std::setw(3) <<k<<std::setfill('0') <<std::setw(3) <<((k&1)?T-t:t)<<".ppm"),true);
      }

      if(repeat){
        //meassure offset
        double off=sqrDistance(W,b[t],hatq[t]);
        //cout <<"off = " <<off <<endl;
        if(false && k>0 && off>.05){
          //cout <<t <<" REPEAT: off = " <<off <<endl;
          t-=dt;
        }
      }
    }


    //evaluate trajectory
    //cout <<"variances over time = ";
    //for(t=0;t<T;t++) cout <<' ' <<trace(B[t]);
    double cost_t,cost1=.0,cost2=.0,length=0.;
    for(t=0;t<T;t++){
      ors->setJointState(b[t]);
      ors->calcNodeFramesFromEdges();
      computeProxiesUsingSwift(*ors,*swift,false);
      //slGetProxies(*ors,*ode);
      updateState(CS);
      if(t>0) cost2 += sqrDistance(W,b[t-1],b[t]);
      if(t>0) length += metricDistance(W,b[t-1],b[t]);
      cost1 += cost_t = getCost(CS,W,t);  //cout <<"cost = " <<cost_t <<endl;
    }
    *os <<std::setw(3) <<k
        <<"  time " <<MT::timerRead(false)
        <<"  cost1 " <<cost1
        <<"  cost2 " <<cost2
        <<"  length " <<length
        <<"  total-cost " <<cost1+cost2 <<endl;
  }

  q = b;
}
*/

#if 0
void SMAC::readCVdef(std::istream& is){
  char c;
  ControlVariable *cv;
  MT::String name,ref1,ref2;
  ors::Frame f;
  uint i,j,k;
  arr mat;
  MT::String::readSkipSymbols=" \n\r\t";
  MT::String::readStopSymbols=" \n\r\t";
  for(;;){
    MT::skip(is);
    is.get(c);
    if(!is.good()) break;
    cv=&CVs.append();
    switch(c){
    case 'p':
      is >>name >>ref1 >>"<" >>f >>">";
      cv->initPos(name,*ors,ors->getName(ref1)->index,f);
      break;
    case 'j':
      is >>name >>ref1;
      cv->initQSingle(name,*ors,ors->getName(ref1)->firstIn->index);
      break;
    case 'l':
      is >>name >>ref1 >>ref2;
      i=ors->getName(ref1)->firstIn->index;
      j=ors->getName(ref2)->firstIn->index;
      mat.resize(j-i+1,ors->getJointStateDimension());
      mat.setZero();
      for(k=0;k<=j-i;k++) mat(k,i+k)=1.;
      cv->initQLinear(name,*ors,mat);
      break;
    case 's':
      is >>name >>ref1 >>ref2;
      i=ors->getName(ref1)->firstIn->index;
      j=ors->getName(ref2)->firstIn->index;
      mat.resize(j-i+1,ors->getJointStateDimension());
      mat.setZero();
      for(k=0;k<=j-i;k++) mat(k,i+k)=1.;
      cv->initQSquared(name,*ors,mat);
      break;
    case 'c':
      is >>name >>ref1;
      cv->initContact(name,*ors,ors->getName(ref1)->index);
      break;
    case 'X':
      is >>name >>ref1;
      cv->initGrip(name,*ors,ors->getName(ref1)->index);
      break;
    case 'o':
      is >>name >>ref1 >>ref2 >>"<" >>f >>">";
      if(ref2=="*")
        cv->initOri(name,*ors,ors->getName(ref1)->index,-1,f);
      else
	cv->initOri(name,*ors,ors->getName(ref1)->index,ors->getName(ref2)->index,f);
      break;
    case 'M':
      is >>name;
      cv->initCom(name,*ors);
      break;
    case 'C':
      is >>name;
      cv->initCollision(name,*ors);
      break;
    default:
      NIY;
    }
    if(is.fail()) HALT("error reading `"<<c<<"' variable in smac");
  }
  is.clear();
}

void SMAC::CVclear(){
  uint i;
  for(i=0;i<CS.N;i++)
    CS(i)->x.clear();
}

void SMAC::plotCVs(){
  plotData.points.resize(2*CS.N);
  uint i;
  for(i=0;i<CS.N;i++){
    plotData.points(2*i) = CS(i)->x;
    plotData.points(2*i+1) = CS(i)->x_target;
  }
}

#endif
